#include "dir.bi"
#include "hash.bi"
#include "crt.bi"

using fb

'' Make sure these are big enough before running.
#define MAXDIRS 2000
#define MAXFILES 10000

#define FALSE 0
#define TRUE 1

type DIRINFO_T
	name as string
end type

type FILEINFO_T
	name as string
end type

enum LISTFILE_FLAGS
	LISTFILE_FLAG_NONE = 0
	LISTFILE_FLAG_NO_VERIFY = 1
end enum

type LISTFILE_T
	name as zstring ptr
	flags as LISTFILE_FLAGS
end type

const NUMLISTFILES = 10
dim shared listfiles(0 to NUMLISTFILES-1) as LISTFILE_T = _
{ _
	( @"dos32.lst", LISTFILE_FLAG_NONE ), _
	( @"win32.lst", LISTFILE_FLAG_NONE ), _
	( @"linux.lst", LISTFILE_FLAG_NONE ), _
	( @"compiler_src.lst", LISTFILE_FLAG_NONE ), _
	( @"rtlib_src.lst", LISTFILE_FLAG_NONE ), _
	( @"gfxlib2_src.lst", LISTFILE_FLAG_NONE ), _
	( @"src-contrib.lst", LISTFILE_FLAG_NONE ), _
	( @"tests.lst", LISTFILE_FLAG_NONE ), _
	( @"manifest.lst", LISTFILE_FLAG_NONE ), _
	( @"ignore.lst", LISTFILE_FLAG_NO_VERIFY ) _
}

/'
	@"src-config.lst", _
	@"src-fbc.lst", _
	@"src-rtlib.lst", _
	@"src-gfxlib.lst", _
'/


#if 1

const def_base_dir = "f:/freebasic/"
const def_list_dir = "current/"

#else

const def_base_dir = "g:/freebasic/"
const def_list_dir = "fbc-0.17/"

#endif

dim shared ndirs as integer
dim shared sdirs( 0 to MAXDIRS) as FILEINFO_T
dim shared dir_hash as HASH

dim shared nfiles as integer
dim shared sfiles( 0 to MAXFILES ) as DIRINFO_T
dim shared file_hash as HASH

dim shared k_nfiles as integer
dim shared k_sfiles( 0 to MAXFILES ) as DIRINFO_T
dim shared k_file_hash as HASH

'':::::
sub add_dir_name( byref n as string )

	if( dir_hash.test( n ) = 0 ) then

		sdirs( nDirs ).name = n
		nDirs += 1
		dir_hash.add( n, @sdirs( nDirs ) )

	end if

end sub

'':::::
function compare_dirs cdecl ( byval a_ as any ptr, byval b_ as any ptr ) as integer
	dim a as DIRINFO_T ptr = a_
	dim b as DIRINFO_T ptr = b_

	if( a->name > b->name ) then
		function = 1
	elseif( a->name < b->name ) then
		function = -1
	else
		function = 0
	end if

end function

'':::::
sub sort_dirs()
	qsort( @sdirs(0), ndirs, sizeof( DIRINFO_T ), @compare_dirs )
end sub

'':::::
function compare_files cdecl ( byval a_ as any ptr, byval b_ as any ptr ) as integer
	dim a as FILEINFO_T ptr = a_
	dim b as FILEINFO_T ptr = b_
	dim i as integer = any
	dim j as integer = any
	
	i = instr( a->name, "/" )
	j = instr( b->name, "/" )

	if(( i > 0 ) and ( j = 0 )) then
		function = 1
	elseif(( i = 0 ) and ( j > 0 )) then
		function = -1
	else
		if( a->name > b->name ) then
			function = 1
		elseif( a->name < b->name ) then
			function = -1
		else
			function = 0
		end if
	end if

end function

'':::::
sub sort_files()
	qsort( @sfiles(0), nfiles, sizeof( FILEINFO_T ), @compare_files )
end sub

'':::::
sub k_sort_files()
	qsort( @k_sfiles(0), k_nfiles, sizeof( FILEINFO_T ), @compare_files )
end sub

'':::::
sub add_file_name( byref n as string )

	if( file_hash.test( n ) = 0 ) then

		sfiles( nFiles ).name = n
		nFiles += 1
		file_hash.add( n, @sfiles( nFiles ) )

	end if

end sub

'':::::
sub add_k_file_name( byref n as string )

	if( k_file_hash.test( n ) = 0 ) then

		k_sfiles( k_nFiles ).name = n
		k_nFiles += 1
		k_file_hash.add( n, @k_sfiles( k_nFiles ) )

	end if

end sub

'':::::
sub scan_dir_for_dirs _
	( _
		byref base_dir as string, _
		byref start_dir as string _
	)

	dim as string d
	dim attrib as integer

	d = dir( base_dir & start_dir & "*.*", fbDirectory, @attrib )
	while( d > "" )
		if( attrib and fbDirectory ) then
			if( d <> "." and d <> ".." ) then
				add_dir_name( start_dir & d & "/" )
			end if
		end if
		d = dir(@attrib)
	wend

end sub

'':::::
sub scan_dirs _
	( _
		byref base_dir as string _
	)

	dim as integer i

	dir_hash.clear()
	ndirs = 0

	add_dir_name( "" )

	while( i < ndirs )
		scan_dir_for_dirs( base_dir, sdirs(i).name )
		i += 1
	wend

end sub

'':::::
sub scan_dir_for_files _
	( _
		byref base_dir as string, _
		byref start_dir as string _
	)

	dim as string d
	dim attrib as integer

	d = dir( base_dir & start_dir & "*.*", fbNormal, @attrib )
	while( d > "" )
		if(( attrib and fbDirectory ) = 0 ) then
			add_file_name( start_dir & d )
		end if
		d = dir(@attrib)
	wend

end sub

'':::::
sub scan_files _
	( _
		byref base_dir as string _
	)
	
	dim i as integer = any

	file_hash.clear()
	nfiles = 0

	for i = 0 to nDirs - 1
		scan_dir_for_files( base_dir, sDirs(i).name )
	next

end sub

'':::::
sub read_list_file( byref f as string )
	dim h as integer, x as string

	h = freefile
	if( open( f for input access read as #h ) = 0 ) then
		while eof(h) = 0
			line input #h, x
			if( trim(x) > "" ) then
				add_k_file_name( x )
			end if
		wend
		close #h
	else
		print "Unable to read '" & f & "'"
	end if

end sub

'':::::
sub read_list_files( byref list_dir as string )
	dim i as integer
	k_file_hash.clear()
	k_nfiles = 0
	print "Reading list files ..."
	for i = 0 to NUMLISTFILES-1
		read_list_file( list_dir & *listfiles( i ).name )
	next i
end sub

'':::::
sub sort_list_file _
	( _
		byref list_dir as string, _
		byref filename as string _
	)
	dim as integer i, h

	k_file_hash.clear()
	k_nfiles = 0
	read_list_file( list_dir & filename )
	k_sort_files()
	h = freefile
	open list_dir & filename for output as #h
	for i = 0 to k_nfiles - 1
		print #h, k_sfiles(i).name
	next i

end sub

sub sort_list_files _
	( _
		byref list_dir as string _
	)

	dim i as integer
	for i = 0 to NUMLISTFILES-1
		print "Sorting '" & list_dir & *listfiles(i).name & "'"
		sort_list_file( list_dir, *listfiles(i).name )
	next i

end sub



sub gen_added_and_deleted _
	( _
		byref base_dir as string, _
		byref list_dir as string _
	)

	dim i as integer, h as integer

	print "Scanning '" & base_dir & "' for directories"
	scan_dirs( base_dir )
	sort_dirs()

	print "Scanning '" & base_dir & "' for files"
	scan_files( base_dir )
	sort_files()

	read_list_files( list_dir )

	print "Generating '" & list_dir & "added.lst'"
	h = freefile
	open list_dir & "added.lst" for output as #h
	for i = 0 to nfiles - 1
		if( k_file_hash.test( sfiles(i).name ) = 0 ) then
			'' added
			print #h, sfiles(i).name
		end if
	next i
	close #h

	print "Generating '" & list_dir & "deleted.lst'"
	open list_dir & "deleted.lst" for output as #h
	for i = 0 to k_nfiles - 1
		if( file_hash.test( k_sfiles(i).name ) = 0 ) then
			'' deleted
			print #h, k_sfiles(i).name
		end if
	next i
	close #h
end sub

'':::::
sub create_list_file _
	( _
		byref base_dir as string, _
		byref list_dir as string, _
		byref list_file as string _
	)

	dim as integer i, h

	print "Scanning '" & base_dir & "' for directories"
	scan_dirs( base_dir )
	sort_dirs()

	print "Scanning '" & base_dir & "' for files"
	scan_files( base_dir )
	sort_files()

	print "Generating '" & list_dir & list_file & "'"
	h = freefile
	open list_dir & list_file for output as #h
	for i = 0 to nfiles - 1
		print #h, sfiles(i).name
	next i
	close #h

end sub

'':::::
sub clean_deleted_files_from_listfile _
	( _
		byref base_dir as string, _
		byref list_dir as string, _
		byref list_file as string _
	)
	
	dim as integer i, h

	k_file_hash.clear()
	k_nfiles = 0

	read_list_file( list_dir & list_file )

	h = freefile
	open list_dir & list_file for output as #h
	for i = 0 to k_nfiles - 1
		if( dir( base_dir & k_sfiles(i).name ) > "" ) then
			print #h, k_sfiles(i).name
		end if
	next i
	close #h

end sub

'':::::
sub clean_deleted_files_from_listfiles _
	( _
		byref base_dir as string, _
		byref list_dir as string _
	)

	dim i as integer
	for i = 0 to NUMLISTFILES-1
		print "Cleaning deleted from '" & list_dir & *listfiles(i).name & "'"
		clean_deleted_files_from_listfile( base_dir, list_dir, *listfiles(i).name )
	next i


end sub

'':::::
sub verify_files_in_listfile _
	( _
		byref base_dir as string, _
		byref list_dir as string, _
		byref list_file as string _
	)
	
	dim as integer i, h

	k_file_hash.clear()
	k_nfiles = 0

	read_list_file( list_dir & list_file )

	for i = 0 to k_nfiles - 1
		if( dir( base_dir & k_sfiles(i).name ) = "" ) then
			print #h, list_file & ": missing : " & k_sfiles(i).name
		end if
	next i

end sub

'':::::
sub verify_files_in_listfiles _
	( _
		byref base_dir as string, _
		byref list_dir as string _
	)

	dim i as integer
	for i = 0 to NUMLISTFILES-1
		if( ( listfiles(i).flags and LISTFILE_FLAG_NO_VERIFY ) = 0 ) then
			verify_files_in_listfile( base_dir, list_dir, *listfiles(i).name )
		end if
	next i


end sub

dim as string base_dir, list_dir, list_file
dim i as integer

select case lcase(command(1))
case "sort"

	if lcase(command(2)) = "" then

		base_dir = def_base_dir
		list_dir = def_list_dir

		sort_list_files( list_dir )

	else
	
		base_dir = def_base_dir
		list_dir = command(2)
		list_file = command(3)

		sort_list_file( list_dir, list_file )

	end if

case "check"
	base_dir = def_base_dir

	if lcase(command(2)) = "" then
		list_dir = def_list_dir
	else
		list_dir = command(2)
	end if

	gen_added_and_deleted( base_dir, list_dir )

case "create"
	base_dir = command(2)
	list_dir = command(3)
	list_file = command(4)

	create_list_file( base_dir, list_dir, list_file )

case "clean"
	base_dir = def_base_dir
	list_dir = command(2)

	if( list_dir = "" ) then
		list_dir = def_list_dir
	end if

	list_file = command(3)

	i = 3
	list_file = command(i)
	if( list_file > "" ) then
		while( list_file > "" )
			clean_deleted_files_from_listfile( base_dir, list_dir, list_file )
			i += 1
			list_file = command(i)
		wend
	else
		clean_deleted_files_from_listfiles( base_dir, list_dir )
	end if

case "verify"
	base_dir = def_base_dir
	list_dir = command(2)

	if( list_dir = "" ) then
		list_dir = def_list_dir
	end if

	list_file = command(3)

	i = 3
	list_file = command(i)
	if( list_file > "" ) then
		while( list_file > "" )
			verify_files_in_listfile( base_dir, list_dir, list_file )
			i += 1
			list_file = command(i)
		wend
	else
		verify_files_in_listfiles( base_dir, list_dir )
	end if

case else
	print "flist check"
	print "flist sort [listdir/ listfile...]"
	print "flist create basedir/ listdir/ listfile"
	print "flist clean [listdir/ listfile...]"
	print "flist verify [listdir/ listfile...]"
	print
	print "note: paths must end with '/'"
	print "      check the source, there is a hard-coded path, sorry."

end select
